/*
 * cblas.hpp
 *
 *  Created on: Dec 4, 2011
 *      Author: raoul
 */

#ifndef CBLAS_HPP_
#define CBLAS_HPP_

//#include <complex>

#ifndef BLAS_LIBRARY
// 1 == ATLAS, 2 == Intel MKL
#define BLAS_LIBRARY 1
#endif

#if BLAS_LIBRARY==1
  extern "C"{
    #include "clapack.h"
  }
#elif BLAS_LIBRARY==2
  extern "C"{
    #include "clapack.h"
  }
  #include "mkl_lapack.h"
#else
  #error Only support for ATLAS and MKL is implemented.
#endif

namespace blas {

typedef CBLAS_INDEX blas_index;
//typedef std::complex<double> blas_double_complex;
//typedef std::complex<float> blas_single_complex;

// --- test ---

// ----- helpers -----

inline CBLAS_TRANSPOSE
trans(const bool & transpose)
{
  return (transpose) ? CblasTrans : CblasNoTrans;
}

// ----- template declarations -----

// reduction
template <typename ValueType>
ValueType asum(const int N, const ValueType *X, const int incX);

template <typename ValueType>
blas_index iamax(const int N, const ValueType *X, const int incX);

// multiplication
/*
 *       K                               N
 *     —————            N            —————————
 *     |   |        —————————        |       |
 *   M | A |   *  K |   B   |   =  M |   C   |
 *     |   |        —————————        |       |
 *     —————                         —————————
 */
template <typename ValueType>
void gemm(const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA,
          const enum CBLAS_TRANSPOSE TransB, const int M, const int N,
          const int K, const ValueType alpha, const ValueType *A,
          const int lda, const ValueType *B, const int ldb,
          const ValueType beta, ValueType *C, const int ldc);

/*
 *
 *       M              N                N
 *     —————        —————————        —————————
 *   M | A |   *  M |   B   |   =  M |   C   |
 *     —————        —————————        —————————
 *
 */
template <typename ValueType>
void symm(const enum CBLAS_ORDER Order, const enum CBLAS_SIDE Side,
          const enum CBLAS_UPLO Uplo, const int M, const int N,
          const ValueType alpha, const ValueType *A, const int lda,
          const ValueType *B, const int ldb, const ValueType beta,
          ValueType *C, const int ldc);

/*
 *       K               N
 *     —————         —————————
 *     |   |         |       |
 *   N | A |       N |   C   |
 *     |   |         |       |
 *     —————         —————————
 */
template <typename ValueType>
void syrk(const enum CBLAS_ORDER Order, const enum CBLAS_UPLO Uplo,
          const enum CBLAS_TRANSPOSE Trans, const int N, const int K,
          const ValueType alpha, const ValueType *A, const int lda,
          const ValueType beta, ValueType *C, const int ldc);

// LU factorization and inversion
template <typename ValueType>
int getrf(const enum CBLAS_ORDER Order, const int M, const int N,
          ValueType *A, const int lda, int *ipiv);

template <typename ValueType>
int getri(const enum CBLAS_ORDER Order, const int N, ValueType *A,
          const int lda, const int *ipiv);

template <typename ValueType>
int potrf(const enum ATLAS_ORDER Order, const enum ATLAS_UPLO Uplo,
          const int N, ValueType *A, const int lda);

template <typename ValueType>
int potri(const enum ATLAS_ORDER Order, const enum ATLAS_UPLO Uplo,
          const int N, ValueType *A, const int lda);

// QR factorization
template <typename ValueType>
int geqrf(const int M, const int N, ValueType *A, const int lda,
          ValueType* tau);

template <typename ValueType>
int orgqr(const int M, const int N, const int K, ValueType *A,
          const int lda, const ValueType* tau);

template <typename ValueType>
int ormqr(const enum CBLAS_SIDE Side, const enum CBLAS_TRANSPOSE Trans,
          const int M, const int N, const int K,
          const ValueType *A, const int lda, const ValueType* tau,
          ValueType *C, const int ldc);


// ----- template specializations

// --- double

template <>
blas_index iamax(const int N, const double *X, const int incX)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    return cblas_idamax(N, X, incX);
  #endif
}

template <>
double asum(const int N, const double *X, const int incX)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    return cblas_dasum(N, X, incX);
  #endif
}

template <>
void gemm(const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA,
          const enum CBLAS_TRANSPOSE TransB, const int M, const int N,
          const int K, const double alpha, const double *A,
          const int lda, const double *B, const int ldb,
          const double beta, double *C, const int ldc)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    cblas_dgemm(Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
  #endif
}

template <>
void symm(const enum CBLAS_ORDER Order, const enum CBLAS_SIDE Side,
          const enum CBLAS_UPLO Uplo, const int M, const int N,
          const double alpha, const double *A, const int lda,
          const double *B, const int ldb, const double beta,
          double *C, const int ldc)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    cblas_dsymm(Order, Side, Uplo, M, N, alpha, A, lda, B, ldb, beta, C, ldc);
  #endif
}

template <>
void syrk(const enum CBLAS_ORDER Order, const enum CBLAS_UPLO Uplo,
          const enum CBLAS_TRANSPOSE Trans, const int N, const int K,
          const double alpha, const double *A, const int lda,
          const double beta, double *C, const int ldc)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    cblas_dsyrk(Order, Uplo, Trans, N, K, alpha, A, lda, beta, C, ldc);
  #endif
}

template <>
int getrf(const enum CBLAS_ORDER Order, const int M, const int N,
          double *A, const int lda, int *ipiv)
{
  #if BLAS_LIBRARY==1
    return clapack_dgetrf(Order, M, N, A, lda, ipiv);
  #elif BLAS_LIBRARY==2
    int info;
    if (Order == CblasRowMajor)
      dgetrf(&N, &M, A, &lda, ipiv, &info);
    else
      dgetrf(&M, &N, A, &lda, ipiv, &info);
    return info;
  #endif
}

template <>
int getri(const enum CBLAS_ORDER Order, const int N, double *A,
          const int lda, const int *ipiv)
{
  #if BLAS_LIBRARY==1
    return clapack_dgetri(Order, N, A, lda, ipiv);
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    double dlwork;
    dgetri(&N, A, &lda, ipiv, &dlwork, &neg_one, &info);
    const int lwork(dlwork);
    boost::scoped_array<double> work(new double[lwork]);
    dgetri(&N, A, &lda, ipiv, work.get(), &lwork, &info);
    return info;
  #endif
}

template <>
int potrf(const enum ATLAS_ORDER Order, const enum ATLAS_UPLO Uplo,
          const int N, double *A, const int lda)
{
  #if BLAS_LIBRARY==1
    return clapack_dpotrf(Order, Uplo, N, A, lda);
  #elif BLAS_LIBRARY==2
    const char uplo = ((Order == CblasRowMajor) == (Uplo == CblasLower)) ? 'U' : 'L';
    int info;
    dpotrf(&uplo, &N, A, &lda, &info);
    return info;
  #endif
}

template <>
int potri(const enum ATLAS_ORDER Order, const enum ATLAS_UPLO Uplo,
          const int N, double *A, const int lda)
{
  #if BLAS_LIBRARY==1
    return clapack_dpotri(Order, Uplo, N, A, lda);
  #elif BLAS_LIBRARY==2
    const char uplo = ((Order == CblasRowMajor) == (Uplo == CblasLower)) ? 'U' : 'L';
    int info;
    dpotri(&uplo, &N, A, &lda, &info);
    return info;
  #endif
}

template <>
int geqrf(const int M, const int N, double *A, const int lda, double* tau)
{
  #if BLAS_LIBRARY==1
    #error  // ATLAS does not include dgeqrf
    return 0;
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    double dlwork;
    dgeqrf(&N, &M, A, &lda, tau, &dlwork, &neg_one, &info);
    const int lwork(dlwork);
    boost::scoped_array<double> work(new double[lwork]);
    dgeqrf(&N, &M, A, &lda, tau, work.get(), &lwork, &info);
    return info;
  #endif
}

template <>
int orgqr(const int M, const int N, const int K, double *A,
          const int lda, const double* tau)
{
  #if BLAS_LIBRARY==1
    #error  // ATLAS does not include dorgqr
    return 0;
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    double dlwork;
    dorgqr(&N, &M, &K, A, &lda, tau, &dlwork, &neg_one, &info);
    const int lwork(dlwork);
    boost::scoped_array<double> work(new double[lwork]);
    dorgqr(&N, &M, &K, A, &lda, tau, work.get(), &lwork, &info);
    return info;
  #endif
}

template <>
int ormqr(const enum CBLAS_SIDE Side, const enum CBLAS_TRANSPOSE Trans,
          const int M, const int N, const int K,
          const double *A, const int lda, const double* tau,
          double *C, const int ldc)
{
  #if BLAS_LIBRARY==1
    #error  // ATLAS does not include dormqr
    return 0;
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    double dlwork;
    const char side = (Side == CblasLeft) ? 'R' : 'L'; // sides are switched to compensate for col-major interpretation
    const char trans = (Trans == CblasTrans) ? 'T' : 'N';
    dormqr(&side, &trans, &N, &M, &K, A, &lda, tau, C, &ldc, &dlwork, &neg_one, &info);
    const int lwork(dlwork);
    boost::scoped_array<double> work(new double[lwork]);
    dormqr(&side, &trans, &N, &M, &K, A, &lda, tau, C, &ldc, work.get(), &lwork, &info);
    return info;
  #endif
}

// --- single

template <>
blas_index iamax(const int N, const float *X, const int incX)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    return cblas_isamax(N, X, incX);
  #endif
}

template <>
float asum(const int N, const float *X, const int incX)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    return cblas_sasum(N, X, incX);
  #endif
}

template <>
void gemm(const enum CBLAS_ORDER Order, const enum CBLAS_TRANSPOSE TransA,
          const enum CBLAS_TRANSPOSE TransB, const int M, const int N,
          const int K, const float alpha, const float *A,
          const int lda, const float *B, const int ldb,
          const float beta, float *C, const int ldc)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    cblas_sgemm(Order, TransA, TransB, M, N, K, alpha, A, lda, B, ldb, beta, C, ldc);
  #endif
}

template <>
void symm(const enum CBLAS_ORDER Order, const enum CBLAS_SIDE Side,
          const enum CBLAS_UPLO Uplo, const int M, const int N,
          const float alpha, const float *A, const int lda,
          const float *B, const int ldb, const float beta,
          float *C, const int ldc)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    cblas_ssymm(Order, Side, Uplo, M, N, alpha, A, lda, B, ldb, beta, C, ldc);
  #endif
}

template <>
void syrk(const enum CBLAS_ORDER Order, const enum CBLAS_UPLO Uplo,
          const enum CBLAS_TRANSPOSE Trans, const int N, const int K,
          const float alpha, const float *A, const int lda,
          const float beta, float *C, const int ldc)
{
  #if BLAS_LIBRARY==1 || BLAS_LIBRARY==2
    cblas_ssyrk(Order, Uplo, Trans, N, K, alpha, A, lda, beta, C, ldc);
  #endif
}

template <>
int getrf(const enum CBLAS_ORDER Order, const int M, const int N,
          float *A, const int lda, int *ipiv)
{
  #if BLAS_LIBRARY==1
    return clapack_sgetrf(Order, M, N, A, lda, ipiv);
  #elif BLAS_LIBRARY==2
    int info;
    if (Order == CblasRowMajor)
      sgetrf(&N, &M, A, &lda, ipiv, &info);
    else
      sgetrf(&M, &N, A, &lda, ipiv, &info);
    return info;
  #endif
}

template <>
int getri(const enum CBLAS_ORDER Order, const int N, float *A,
          const int lda, const int *ipiv)
{
  #if BLAS_LIBRARY==1
    return clapack_sgetri(Order, N, A, lda, ipiv);
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    float slwork;
    sgetri(&N, A, &lda, ipiv, &slwork, &neg_one, &info);
    const int lwork(slwork);
    boost::scoped_array<float> work(new float[lwork]);
    sgetri(&N, A, &lda, ipiv, work.get(), &lwork, &info);
    return info;
  #endif
}

template <>
int potrf(const enum ATLAS_ORDER Order, const enum ATLAS_UPLO Uplo,
          const int N, float *A, const int lda)
{
  #if BLAS_LIBRARY==1
    return clapack_spotrf(Order, Uplo, N, A, lda);
  #elif BLAS_LIBRARY==2
    const char uplo = ((Order == CblasRowMajor) == (Uplo == CblasLower)) ? 'U' : 'L';
    int info;
    spotrf(&uplo, &N, A, &lda, &info);
    return info;
  #endif
}

template <>
int potri(const enum ATLAS_ORDER Order, const enum ATLAS_UPLO Uplo,
          const int N, float *A, const int lda)
{
  #if BLAS_LIBRARY==1
    return clapack_spotri(Order, Uplo, N, A, lda);
  #elif BLAS_LIBRARY==2
    const char uplo = ((Order == CblasRowMajor) == (Uplo == CblasLower)) ? 'U' : 'L';
    int info;
    spotri(&uplo, &N, A, &lda, &info);
    return info;
  #endif
}

template <>
int geqrf(const int M, const int N, float *A, const int lda, float* tau)
{
  #if BLAS_LIBRARY==1
    #error  // ATLAS does not include dgeqrf
    return 0;
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    float slwork;
    sgeqrf(&N, &M, A, &lda, tau, &slwork, &neg_one, &info);
    const int lwork(slwork);
    boost::scoped_array<float> work(new float[lwork]);
    sgeqrf(&N, &M, A, &lda, tau, work.get(), &lwork, &info);
    return info;
  #endif
}

template <>
int orgqr(const int M, const int N, const int K, float *A,
          const int lda, const float* tau)
{
  #if BLAS_LIBRARY==1
    #error  // ATLAS does not include dorgqr
    return 0;
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    float slwork;
    sorgqr(&N, &M, &K, A, &lda, tau, &slwork, &neg_one, &info);
    const int lwork(slwork);
    boost::scoped_array<float> work(new float[lwork]);
    sorgqr(&N, &M, &K, A, &lda, tau, work.get(), &lwork, &info);
    return info;
  #endif
}

template <>
int ormqr(const enum CBLAS_SIDE Side, const enum CBLAS_TRANSPOSE Trans,
          const int M, const int N, const int K,
          const float *A, const int lda, const float* tau,
          float *C, const int ldc)
{
  #if BLAS_LIBRARY==1
    #error  // ATLAS does not include dormqr
    return 0;
  #elif BLAS_LIBRARY==2
    int info;
    const int neg_one(-1);
    float slwork;
    const char side = (Side == CblasLeft) ? 'R' : 'L'; // sides are switched to compensate for col-major interpretation
    const char trans = (Trans == CblasTrans) ? 'T' : 'N';
    sormqr(&side, &trans, &N, &M, &K, A, &lda, tau, C, &ldc, &slwork, &neg_one, &info);
    const int lwork(slwork);
    boost::scoped_array<float> work(new float[lwork]);
    sormqr(&side, &trans, &N, &M, &K, A, &lda, tau, C, &ldc, work.get(), &lwork, &info);
    return info;
  #endif
}

} /* namespace blas */

#endif /* CBLAS_HPP_ */







