/*
 * matrix_generator.hpp
 *
 *  Created on: Dec 17, 2011
 *      Author: raoul
 */

#ifndef MATRIX_GENERATOR_HPP_
#define MATRIX_GENERATOR_HPP_

#include <ctime>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/lognormal_distribution.hpp>
#include <boost/random/uniform_real_distribution.hpp>
#include <boost/random/exponential_distribution.hpp>

#include "matrix.hpp"

template <typename ElementType>
struct matrix_generator
{
  typedef matrix<ElementType> matrix_type;
  typedef typename matrix_type::element_type element_type;
  typedef typename matrix_type::size_type size_type;

  class inverse_diagonal_ascending
  {
    const size_type size;

  public:
    inverse_diagonal_ascending(const size_type size)
      : size(size)
    {}

    static const bool produces_symmetry = false;

    inline element_type
    operator () (const size_type row, const size_type col)
    {
      return (row + col == size - 1) ? row + 1 : 0;
    }
  };

  class diagonal_ascending
  {
  public:
    static const bool produces_symmetry = true;

    inline element_type
    operator () (const size_type row, const size_type col)
    {
      return (row == col) ? row + 1 : 0;
    }
  };

  class inverse_tridiagonal_ones
  {
    const size_type size;

  public:
    inverse_tridiagonal_ones(const size_type size)
      : size(size)
    {}

    static const bool produces_symmetry = true;

    inline element_type
    operator () (const size_type row, const size_type col)
    {
      return (row + col <= size - 1) ? 1 : 0;
    }
  };

  class normal_distribution
  {
    boost::random::mt19937_64 gen;
    boost::random::normal_distribution<element_type> dist;

  public:
    normal_distribution(const element_type mean, const element_type sigma, time_t & seed)
      : gen(seed = (seed == 0) ? std::time(0) : seed),
        dist(mean, sigma)
    {}

    static const bool produces_symmetry = false;

    inline element_type
    operator () (const size_type, const size_type)
    {
      return dist(gen);
    }
  };

  template <typename Generator>
  static inline matrix_type
  generate(Generator gen, const size_type size)
  {
    return generate(gen, size, size);
  }

  template <typename Generator>
  static matrix_type
  generate(Generator gen, const size_type height, const size_type width)
  {
    matrix_type M(height, width);
    // possible parallelization: parallel random generation
    for (size_type row = 0; row < M.get_height(); ++row)
      for (size_type col = 0; col < M.get_width(); ++col)
        M.elem(row, col) = gen(row, col);
    M.set_symmetric(gen.produces_symmetry);
    return M;
  }

  class dist_uni_div_uni
  {
    boost::random::uniform_real_distribution<element_type> dist;

  public:
    explicit
    dist_uni_div_uni(element_type max_cond)
      : dist(1, sqrt(max_cond))
    {}

    template <typename Generator>
    element_type
    operator () (Generator & gen)
    {
      element_type X = dist(gen),
                   Y = dist(gen);
      return X / Y;
    }
  };

  class dist_uni_div
  {
    boost::random::uniform_real_distribution<element_type> dist;
    element_type d;

  public:
    dist_uni_div(element_type max_cond, element_type divisor)
      : dist(1, max_cond),
        d(divisor)
    {}

    template <typename Generator>
    element_type
    operator () (Generator & gen)
    {
      return dist(gen) / d;
    }
  };

  class dist_div_uni
  {
    boost::random::uniform_real_distribution<element_type> dist;

  public:
    dist_div_uni(element_type max_cond)
      : dist(1, max_cond)
    {}

    template <typename Generator>
    element_type
    operator () (Generator & gen)
    {
      return element_type(1) / dist(gen);
    }
  };

  class dist_loguni
  {
    boost::random::uniform_real_distribution<element_type> dist;

  public:
    dist_loguni(element_type max_log_cond)
      : dist(-(max_log_cond / 2), max_log_cond / 2)
    {}

    dist_loguni(element_type min_log_cond, element_type max_log_cond)
      : dist(min_log_cond, max_log_cond)
    {}

    template <typename Generator>
    element_type
    operator () (Generator & gen)
    {
      return pow(2, dist(gen));
    }
  };

  static matrix_type
  spd_by_cholesky_decomposition(const size_type size, const element_type sigma_diag,
      const element_type sigma_non_diag, time_t & seed)
  {
    matrix_type M(size);
    boost::random::mt19937_64 gen(seed = ((seed == 0) ? std::time(0) : seed));
    boost::random::exponential_distribution<element_type> diagonal(sigma_diag);
    boost::random::normal_distribution<element_type> non_diag(0, sigma_non_diag);

    // generate an upper triangular matrix
    for (size_type row = 0; row < M.get_height(); ++row)
    {
      for (size_type col = 0; col < row; ++col)
        M.elem(row, col) = 0;
      M.elem(row, row) = 1 + diagonal(gen);
      for (size_type col = row + 1; col < M.get_width(); ++col)
        M.elem(row, col) = non_diag(gen);
    }

    return M.spd_square(1, true);
  }

  // Generates a symmetric positive definite matrix.
  //   Its eigenvalues are explicitly generated by eigenvalue.
  //   A matrix with those eigenvalues generated by multiplying with a uniformly chosen orthogonal matrix.
  template <typename Distribution>
  static matrix_type
  spd_by_eigendecomposition(const size_type size, Distribution eigenvalue, time_t & seed)
  {
    boost::random::mt19937_64 gen(seed = ((seed == 0) ? std::time(0) : seed));
    boost::random::normal_distribution<element_type> normal(0, 1);

    // generate normally distributed matrix entries
    matrix_type Q(size);
    for (size_type row = 0; row < size; ++row)
      for (size_type col = 0; col < size; ++col)
        Q.elem(row, col) = normal(gen);
    {
      // compute QR factorization to get a uniformly distributed orthonormal matrix
      boost::scoped_array<element_type> tau(new element_type[size]);
      blas::geqrf(size, size, Q.get_pointer(), size, tau.get());
      // extract Q and signs of Diag(R)
      boost::scoped_array<int> signs_R(new int[size]);
      #pragma omp parallel for
      for (size_type col = 0; col < size; ++col)
        if (Q.const_elem(col,col) >= 0)
          signs_R[col] = 1;
        else
          signs_R[col] = -1;
      blas::orgqr(size, size, size, Q.get_pointer(), size, tau.get());
      // transform to Q of QR factorization positive elements on Diag(R). This Q is uniformly distributed.
      #pragma omp parallel for
      for (size_type col = 0; col < size; ++col)
        for (size_type row = 0; row < size; ++row)
          // LAPACK stores Q in column-major, so the indices are swapped
          Q.elem(col,row) *= signs_R[col];
    }
    // At this point, Q is being logically transposed, thus column-major becomes row-major.
    // That makes no difference, since (Q -> Q^T) is a bijection and Q is uniformly distributed
    //   from all orthogonal matrices.

    matrix_type DQ(size);
    DQ.make_copy_of(Q);

    // multiply O with random eigenvalues, i.e. D * O
    //element_type max_ev = 0, min_ev = std::numeric_limits<element_type>::max();
    for (size_type row = 0; row < size; ++row)
    {
      element_type l = eigenvalue(gen);
      //if (max_ev < l)
        //max_ev = l;
      //if (min_ev > l)
        //min_ev = l;
      for (size_type col = 0; col < size; ++col)
        DQ.elem(row, col) *= l;
    }
    //VERBOSE("min_ev = " << min_ev << " max_ev = " << max_ev << "\n");
    // compute symmetric positive definite from eigendecomposition O^T * D * O
    matrix_type QtDQ(size);
    QtDQ.multiply_symmetric(Q, DQ, 1, 0, true, false);
    return QtDQ;
  }
};

#endif /* MATRIX_GENERATOR_HPP_ */









