/*
 * matrix_inverse.hpp
 * for inclusion in matrix.hpp
 *
 *  Created on: Nov 28, 2011
 *      Author: raoul
 */

#include <math.h>
#include <limits>

#include "recursive.hpp"

template <typename ElementType>
class matrix_inverse
{
  /* available base inversion methods:
   *   L: call LAPACK inversion
   *   N: Newton inversion
   *   d: set the diagonal to 1/one_norm(M) as initially for Newton inversion
   * available recursive inversion methods:
   *   n: one iteration of Newton inversion (to improve the approximation)
   *   S: Strassen inversion
   */

  typedef matrix<ElementType> matrix_type;
  typedef typename matrix_type::element_type element_type;
  typedef typename matrix_type::size_type size_type;
  typedef typename matrix_type::partition partition_type;

  matrix_type I;

public:
  operator matrix_type & ()
  {
    return I;
  }

  matrix_inverse(const matrix_type & M,
      const bool is_symmetric_positive_definite = false,
      const std::string method = "L")
    : I()
  {
    assert(M.get_height() == M.get_width());
    assert(method.length() > 0);

    if (method == "L")
    {
      I = matrix_type(M.get_height());
      if (is_symmetric_positive_definite)
        lapack_inversion<true>::invert(M, I);
      else
        lapack_inversion<false>::invert(M, I);
    }
    else
    {
      matrix_type S,
                  Si(M.get_height());
      if (is_symmetric_positive_definite)
      {
        S = M;
      }
      else
      {
        // calculate symmetric positive definite M*M^T
        S = M.spd_square();
      }

      // invert S

      typedef recursion<const matrix_type &, matrix_type &> recursion_type;
      recursion_type r;
      {
        for (std::string::const_iterator it = method.begin(); it != method.end(); ++it)
          if (*it == 'S')
            r.add_function(strassen_inversion);
          else if (*it == 'n')
            if (it + 1 == method.end() || *(it + 1) < '0' || *(it + 1) > '5')
              r.add_function(newton_stabilization<>::run);
            else
            {
              ++it;
              if (*it == '0')
                r.add_function(newton_stabilization<0>::run);
              else if (*it == '1')
                r.add_function(newton_stabilization<1>::run);
              else if (*it == '2')
                r.add_function(newton_stabilization<2>::run);
              else if (*it == '3')
                r.add_function(newton_stabilization<3>::run);
              else if (*it == '4')
                r.add_function(newton_stabilization<4>::run);
              else if (*it == '5')
                r.add_function(newton_stabilization<5>::run);
            }
          else if (*it == 'N')
            r.add_function(newton_inversion_spd);
          else if (*it == 'd')
            r.add_function(diagonal_inversion_approximation);
          else if (*it == 'L' || *it == 'l')
            r.add_function(lapack_inversion<true>::invert);
          else
            throw typename matrix_type::unknown_inversion_method_error(*it);
      }

      r(S, Si);

      if (*(method.begin()) == 'S' || *(method.begin()) == 'N' || *(method.begin()) == 'n')
        Si.make_symmetric();

      if (is_symmetric_positive_definite)
        I = Si;
      else
      {
        // calculate M^-1 = M^T*(M^-T*M^-1) = M^T*(M*M^T)^-1 = M^T*S^-1
        I = matrix_type(M.get_height());
        if (M.is_symmetric())
          I.multiply_symmetric(M, Si, 1, 0, true, false);
        else
          I.multiply(M, Si, 1, 0, true, false);
      }
    }
  }

private:
  // set I to M^-1
  // M and I the same is allowed (results in inplace operation)
  // assumes M and I square
  // asserts M and I of same height and width
  template <bool is_symmetric_positive_definite>
  struct lapack_inversion
  {
    template <typename Recursion>
    inline static void
    invert(const matrix_type & M, matrix_type & I, Recursion)
    {
      invert(M, I);
    }

    static void
    invert(const matrix_type & M, matrix_type & I)
    {
      #ifdef MATRIX_DO_STAT
        m_stat.start_timer();
      #endif
      I.make_copy_of(M);

      if (is_symmetric_positive_definite)
      {
        int info; // info != 0 indicates a singular matrix
	#ifdef MATRIX_DO_STAT
          m_stat.start_timer();
        #endif
	#ifdef MATRIX_DO_STAT
          m_stat.start_timer();
        #endif
        info = blas::potrf(CblasRowMajor, CblasLower, I.get_height(), I.get_pointer(), I.get_row_length());
        if (info)
          throw typename matrix_type::singular_inversion_error();
        #ifdef MATRIX_DO_STAT
          m_stat.stop_timer();
        #endif
        info = blas::potri(CblasRowMajor, CblasLower, I.get_height(), I.get_pointer(), I.get_row_length());
        if (info)
          throw typename matrix_type::singular_inversion_error();
        #ifdef MATRIX_DO_STAT
          m_stat.stop_timer();
        #endif
        I.make_symmetric();
      }
      else
      {
        boost::scoped_array<int> P(new int [I.get_height()]);
        int info; // info != 0 indicates a singular matrix
        info = blas::getrf(CblasRowMajor, I.get_height(), I.get_width(), I.get_pointer(), I.get_row_length(), P.get());
        if (info)
          throw typename matrix_type::singular_inversion_error();
        info = blas::getri(CblasRowMajor, I.get_height(), I.get_pointer(), I.get_row_length(), P.get());
        if (info)
          throw typename matrix_type::singular_inversion_error();
        I.set_symmetric(M.is_symmetric());
      }
      #ifdef MATRIX_DO_STAT
        m_stat.stop_timer();
      #endif
    }
  };

  template <typename Recursion>
  inline static void
  newton_inversion_spd(const matrix_type & M, matrix_type & I, Recursion)
  {
    newton_inversion_spd(M, I);
  }

  // requires M symmetric positive definite
  // assumes M and I square and same size
  // M and I the same is possible
  // if (EXPLOIT_SYMMETRIC_MULTIPLICATION) I will be symmetric, so only the lower triangle is guaranteed to be set
  static void
  newton_inversion_spd(const matrix_type & M, matrix_type & I)
  {
    #ifdef MATRIX_DO_STAT
      m_stat.start_timer();
    #endif
    #if COMPACT_BEFORE_NEWTON
      matrix_type Mr = M;
      Mr.compact();
    #else
      const matrix_type & Mr = M;
    #endif

    const size_type & size = Mr.get_height();

    // initial approximation
    matrix_type Y_new(size);
    diagonal_inversion_approximation(Mr, Y_new);
    // compute initial error matrix (first multiplication of Newton iteration) and error-measure
    matrix_type E(size);
    E.set_alpha_unity(2);
    E.multiply(Y_new, Mr, -1, 1);
    element_type error_new = E.max_norm_diff_unity();

    matrix_type Y(size);
    element_type error = (std::numeric_limits<element_type>::has_infinity)
            ? std::numeric_limits<element_type>::infinity() : std::numeric_limits<element_type>::max();
    int iteration_count = -1;
    // possible improvement: use 5th and 6th power
    while (error_new < error ||
        (error_new < error && error > MAXIMAL_NEWTON_ERROR))
    {
      // accept improvement
      
      //std::cout << "Newton-Schritt: " << iteration_count << " Error-Alt: " << error << " Error-Neu: " << error_new << "\n";
      //std::cout << "Exponenten: " << pow(error_new,MINIMAL_NEWTON_IMPROVEMENT_NUMERATOR) <<"  "<< pow(error, MINIMAL_NEWTON_IMPROVEMENT_DENOMINATOR) << "\n";
      //std::cout << "\n";
      
      Y.swap(Y_new);
      error = error_new;
      //VERBOSE(error << "\n");
      ++iteration_count;
      if (iteration_count > MAX_NEWTON_ITERATIONS)
        throw typename matrix_type::bad_condition_inversion_error(iteration_count, error);
      // +++ invariant: Y holds the currently known best approximation, error its error-measure.
      // try to improve approximation (second multiplication of Newton iteration)
      Y_new.multiply_symmetric(E, Y, 1, 0, false, false, true);
      // compute error matrix (first multiplication of Newton iteration) and error-measure
      E.set_alpha_unity(2);
      E.multiply(Y_new, Mr, -1, 1);
      error_new = E.max_norm_diff_unity();
    }
    if (error_new < error)
    {
      // accept improvement
      Y.swap(Y_new);
      error = error_new;
      //VERBOSE(error << "\n");
      ++iteration_count;
    }
    PRINT_ERROR_TREE(error << " " << iteration_count);
    I.make_copy_of(Y);

    #ifdef MATRIX_DO_STAT
      m_stat.stop_timer();
    #endif
  }
  
  template <typename Recursion>
  inline static void
  diagonal_inversion_approximation(const matrix_type & M, matrix_type & I, Recursion)
  {
    diagonal_inversion_approximation(M, I);
  }

  // assumes M and I square and same size
  // M and I the same is possible
  inline static void
  diagonal_inversion_approximation(const matrix_type & M, matrix_type & I)
  {
    element_type t = M.one_norm();
    if (t == 0)
      throw typename matrix_type::singular_inversion_error();
    t = element_type(1) / t;
    I.set_alpha_unity(t);
  }

  template <int Version = 0>
  struct newton_stabilization
  {
    // performs one iteration of Newton inversion
    // sets Y_new to an improved approximation of M^-1, using the approximation Y
    // T is used for temporary storage and not read
    // Y_new may be M, or Y may be M, but not both
    // if Y is symmetric, only the lower triangle has to be set
    // if (EXPLOIT_SYMMETRIC_MULTIPLICATION) if I will be symmetric, only the lower triangle is guaranteed to be set
    inline static void
    newton_inversion_iteration(const matrix_type & M, matrix_type & Y_new, matrix_type & Y,
        matrix_type & T)
    {
      if (Version == 0)  // (2I-YM)Y
      {
        T.set_alpha_unity(2);
        T.multiply(Y, M, -1, 1);
        PRINT_ERROR_TREE(T.max_norm_diff_unity());
        if (M.is_symmetric() && Y.is_symmetric())
          Y_new.multiply_symmetric(T, Y, 1, 0, false, false, true);
        else
          Y_new.multiply(T, Y);
      }
      else if (Version == 1)  // Y(2I-MY), Y has to be complete
      {
        T.set_alpha_unity(2);
        T.multiply(M, Y, -1, 1);
        PRINT_ERROR_TREE(T.max_norm_diff_unity());
        if (M.is_symmetric() && Y.is_symmetric())
          Y_new.multiply_symmetric(Y, T, 1, 0, false, false, true);
        else
          Y_new.multiply(Y, T);
      }
      else if (Version == 2)  // Y+(I-YM)Y
      {
        T.set_alpha_unity(1);
        T.multiply(Y, M, -1, 1);
        PRINT_ERROR_TREE(T.max_norm());
        if (M.is_symmetric() && Y.is_symmetric())
          Y_new.multiply_symmetric(T, Y, 1, 0, false, false, true);
        else
          Y_new.multiply(T, Y);
        Y_new += Y;
      }
      else if (Version == 3)  // Y+Y(I-MY), Y has to be complete
      {
        T.set_alpha_unity(1);
        T.multiply(M, Y, -1, 1);
        PRINT_ERROR_TREE(T.max_norm());
        if (M.is_symmetric() && Y.is_symmetric())
          Y_new.multiply_symmetric(Y, T, 1, 0, false, false, true);
        else
          Y_new.multiply(Y, T);
        Y_new += Y;
      }
      else if (Version == 4)  // 2Y-(YM)Y
      {
        T.multiply(Y, M);
        PRINT_ERROR_TREE(T.max_norm_diff_unity());
        if (M.is_symmetric() && Y.is_symmetric())
          Y_new.multiply_symmetric(T, Y, 1, 0, false, false, true);
        else
          Y_new.multiply(T, Y);
        Y_new.add(Y, 2, -1);
      }
      else if (Version == 5)  // 2Y-Y(MY), Y has to be complete
      {
        T.multiply(M, Y);
        PRINT_ERROR_TREE(T.max_norm_diff_unity());
        if (M.is_symmetric() && Y.is_symmetric())
          Y_new.multiply_symmetric(Y, T, 1, 0, false, false, true);
        else
          Y_new.multiply(Y, T);
        Y_new.add(Y, 2, -1);
      }
    }

    template <typename Recursion>
    static void
    run(const matrix_type & M, matrix_type & I, Recursion recurse)
    {
      #ifdef MATRIX_DO_STAT
        m_stat.start_timer();
      #endif
      matrix_type Y(M.get_height());
      // compute numerically less stable inverse
      recurse(M, Y);
      // improve the approximation with one Newton iteration
      matrix_type T(M.get_height());
      newton_inversion_iteration(M, I, Y, T);
      #ifdef MATRIX_DO_STAT
        m_stat.stop_timer();
      #endif
    }
  };

  // sets I to M^-1
  // requires M symmetric positive definite
  // assumes M and I square and same size
  // M and I the same is possible
  // I will be symmetric, so only the lower triangle is guaranteed to be set
  template <typename Recursion>
  static void
  strassen_inversion(const matrix_type & M, matrix_type & I, Recursion recurse)
  {
    #ifdef MATRIX_DO_STAT
      m_stat.start_timer();
    #endif
    // partition
    const partition_type pM(M);
    // M up-left square
    const matrix_type & B = pM.ul;
    // M down-left, possibly not square
    const matrix_type & C = pM.dl;
    // M down-right square
    const matrix_type & D = pM.dr;

    partition_type pI(I);
    // I up-left square
    matrix_type & R = pI.ul;
    // I down-left, possibly not square
    matrix_type & mSiCBi = pI.dl;
    // I down-right square
    matrix_type & Si = pI.dr;

    // invert using recursion
    matrix_type & Bi = R;
    recurse(B, Bi);
    matrix_type CBi = C * Bi;
    { // to delete S (misleading name)
      matrix_type & S = Si;
      S.make_copy_of(D);
      S.multiply_symmetric(CBi, C, -1, 1, false, true);
      recurse(S, Si);
    }
    mSiCBi.multiply(Si, CBi, -1);
    // currently R = Bi
    R.multiply_symmetric(CBi, mSiCBi, -1, 1, true, false, true);

    I.set_symmetric(true);
    #ifdef MATRIX_DO_STAT
      m_stat.stop_timer();
    #endif
  }
};









