/*
 * utilities.hpp
 *
 *  Created on: Jan 27, 2012
 *      Author: raoul
 */

#ifndef UTILITIES_HPP_
#define UTILITIES_HPP_

#include <stdlib.h>
#include <iostream>
#include <cstring>
#include <time.h>
#include <list>

// --- debugging ---

#ifndef VERBOSE
#define VERBOSE(x) std::cout << x
#endif

// --- meta programming ---

void uninstantiable();

class empty_class {};

inline static void null_function() {}

// --- memory ---

template <typename ValueType>
void memcopy(ValueType * to, const ValueType * from, const size_t n)
{
  memcpy(to, from, n * sizeof(ValueType));
}

// --- math ---

template <typename ValueType>
inline ValueType absolute(const ValueType n)
{
  if (n < 0)
    return -n;
  else
    return n;
}

template <typename ValueType>
inline bool absolute_greater(const ValueType a, const ValueType b)
{
  if (a < 0)
    if (b < 0)
      return -a > -b;
    else
      return false;
  else
    if (b < 0)
      return true;
    else
      return a > b;
}

// --- timing ---

template <long TicsPerSecond = 1000000> // needs to be a divisor of 10^9
class timer
{
public:
  static const long tics_per_sec = TicsPerSecond;

private:
  long time;

public:
  timer()
    : time(0)
  {}

  inline void
  start()
  {
    stop();
  }

  inline void
  stop()
  {
    timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    time=ts.tv_sec * TicsPerSecond + ts.tv_nsec / (1000000000 / TicsPerSecond) - time;
  }

  inline void
  reset()
  {
    time = 0;
  }

  inline long
  get_tics() const
  {
    return time;
  }

  inline double
  get_seconds() const
  {
    return double(time) / TicsPerSecond;
  }
};

template <long TicsPerSecond = 1000000> // needs to be a divisor of 10^9
class nested_timer
{
public:
  typedef std::list<timer<TicsPerSecond> > times_container_type;

  static const long tics_per_sec = TicsPerSecond;

private:
  typedef typename times_container_type::iterator times_container_iterator_type;

  times_container_type times;
  times_container_iterator_type next;

public:
  nested_timer()
    : times(1),
      next(times.begin())
  {}

  inline void
  start()
  {
    if (next == times.end())
    {
      times.push_back(timer<TicsPerSecond>());
      next = times.end();
      times.back().start();
    }
    else
    {
      next->start();
      ++next;
    }
  }

  inline void
  stop()
  {
    --next;
    next->stop();
  }

  inline void
  reset()
  {
    times.resize(1);
    next = times.begin();
    next->reset();
  }

  std::string
  get_tics() const
  {
    std::ostringstream oss;
    if (next != times.end())
    {
      oss << next->get_tics();
      times_container_iterator_type it = next;
      ++it;
      for (; it != times.end(); ++it)
        oss << "," << it->get_tics();
    }
    else
      oss << "-";
    return oss.str();
  }

  std::string
  get_seconds() const
  {
    std::ostringstream oss;
    if (next != times.end())
    {
      oss << next->get_seconds();
      times_container_iterator_type it = next;
      ++it;
      for (; it != times.end(); ++it)
        oss << "," << it->get_seconds();
    }
    else
      oss << "-";
    return oss.str();
  }

  const times_container_type &
  get_times() const
  {
    return times;
  }

  const timer<TicsPerSecond> &
  front() const
  {
    return times.front();
  }
};

#endif /* UTILITIES_HPP_ */
