/*
 * matrix.hpp
 *
 *  Created on: Dec 03, 2011
 *      Author: raoul
 */

#ifndef MATRIX_HPP_
#define MATRIX_HPP_

#include <fstream>
#include <cstring>
#include <sstream>
#include <cmath>
#include <omp.h>

#include <boost/shared_array.hpp>
#include <boost/scoped_array.hpp>

#include "utilities.hpp"
#include "cblas.hpp"

#ifdef _OPENMP
  #define PARALLEL true
  #include "matrix.tuning.p.hpp"
#else
  #define PARALLEL false
  #include "matrix.tuning.s.hpp"
#endif

#ifndef RECURSIVE_TRANSPOSE_SIZE
  #define RECURSIVE_TRANSPOSE_SIZE pow(2, 16)
#endif

#ifndef PARALLEL_TRANSPOSE_SIZE
  #define PARALLEL_TRANSPOSE_SIZE pow(2, 16)
#endif

#ifndef PARALLEL_COPY_SIZE
  #define PARALLEL_COPY_SIZE pow(2, 16)
#endif

#ifndef PARALLEL_SET_SIZE
  #define PARALLEL_SET_SIZE pow(2, 16)
#endif

#ifndef PARALLEL_ADD_REDUCTION_SIZE
  #define PARALLEL_ADD_REDUCTION_SIZE pow(2, 16)
#endif

#ifndef PARALLEL_MAX_REDUCTION_SIZE
  #define PARALLEL_MAX_REDUCTION_SIZE pow(2, 16)
#endif

#ifndef PARALLEL_DIFF_MAX_REDUCTION_SIZE
  #define PARALLEL_DIFF_MAX_REDUCTION_SIZE pow(2, 16)
#endif

#ifndef EXPLOIT_SYMMETRIC_MULTIPLICATION
  // Warning: The use of OpenMP seems to interfere with BLAS multiplication,
    // leading to unpredictable results (tested with Intel MKL of Composer XE 2011).
  #define EXPLOIT_SYMMETRIC_MULTIPLICATION !PARALLEL
#endif


#if EXPLOIT_SYMMETRIC_MULTIPLICATION
  #ifndef RECURSIVE_MULTIPLY_SIZE
  #define RECURSIVE_MULTIPLY_SIZE pow(2, 16)
  #endif

  #if PARALLEL
    #ifndef PARALLEL_MULTIPLY_SIZE
    #define PARALLEL_MULTIPLY_SIZE pow(2, 16)
    #endif
  #endif

  #ifndef AVOID_STRANGE_MULTIPLICATION_BUG
  #define AVOID_STRANGE_MULTIPLICATION_BUG true
  #endif
#endif

#ifndef COMPACT_BEFORE_NEWTON
#define COMPACT_BEFORE_NEWTON false
#endif

#ifndef MAX_NEWTON_ITERATIONS
#define MAX_NEWTON_ITERATIONS 1024 + 8 * log(size)
#endif

#ifndef MINIMAL_NEWTON_IMPROVEMENT_NUMERATOR
#define MINIMAL_NEWTON_IMPROVEMENT_NUMERATOR 5
#endif

#ifndef MINIMAL_NEWTON_IMPROVEMENT_DENOMINATOR
#define MINIMAL_NEWTON_IMPROVEMENT_DENOMINATOR 6
#endif

#ifndef MAXIMAL_NEWTON_ERROR
#define MAXIMAL_NEWTON_ERROR pow(2, -(2*std::numeric_limits<element_type>::digits / 3))
#endif

#ifndef PRINT_ERROR_TREE
  #define PRINT_ERROR_TREE(x)
#else
  #undef PRINT_ERROR_TREE
  #define PRINT_ERROR_TREE(x) std::cout << x << "\n"
#endif

template <typename ElementType>
class matrix_inverse;

template <typename SizeType>
inline
std::pair<SizeType, SizeType>
partition_size(const SizeType n)
{
  SizeType i = n/2;
  return std::make_pair(n - i, i);
}

template<typename ElementType>
class matrix_base
{
public:
  typedef ElementType element_type;
  typedef blas::blas_index size_type;

private:
  size_type height,
            width;
  bool symmetric;

public:
  // ----- constructors and assignment -----

  explicit
  matrix_base(size_type sidelength = 0, bool symmetric = false)
    : height(sidelength),
      width(sidelength),
      symmetric(symmetric)
  {}

  explicit
  matrix_base(size_type height, size_type width, bool symmetric = false)
    : height(height),
      width(width),
      symmetric(symmetric)
  {}

  ~matrix_base()
  {}

  // ----- informative -----

  inline const size_type &
  get_height() const
  {
    return height;
  }

  inline const size_type &
  get_width() const
  {
    return width;
  }

  inline const bool &
  is_symmetric() const
  {
    return symmetric;
  }

  inline bool
  operator == (const matrix_base & M) const
  {
    return height == M.height && width == M.width && symmetric == M.symmetric;
  }

  // ----- manipulating -----

  inline void
  set_size(const size_type new_size)
  {
    height = width = new_size;
  }

  inline void
  set_size(const size_type new_height, const size_type new_width)
  {
    height = new_height;
    width = new_width;
  }

  inline const bool &
  set_symmetric(const bool s)
  {
    return symmetric = s;
  }

  inline matrix_base &
  swap(matrix_base & M)
  {
    std::swap(height, M.height);
    std::swap(width, M.width);
    std::swap(symmetric, M.symmetric);
    return *this;
  }
};

template<typename ElementType> class matrix;

template <typename ElementType>
std::ostream &
operator << (std::ostream & strm, const matrix<ElementType> & M)
{
  for (typename matrix<ElementType>::size_type row = 0; row < M.get_height(); ++row)
  {
    for (const ElementType * p = M.get_const_pointer(row);
        p < M.get_const_pointer(row) + M.get_width(); ++p)
    {
      const ElementType v = (-0.001 < *p && *p < 0.001) ? 0 : *p;
      strm << v << " ";
    }
    strm << "\n";
  }
  return strm;
}

template<typename ElementType>
class matrix : public matrix_base<ElementType>
{
  typedef matrix_base<ElementType> matrix_b;
public:
  typedef typename matrix_b::element_type element_type;
  typedef typename matrix_b::size_type size_type;
protected:
  typedef boost::shared_array<element_type> matrix_data_type;

public:
  struct singular_inversion_error
  {};

  struct bad_condition_inversion_error
  {
    const int iterations;
    const element_type remaining_error;

    bad_condition_inversion_error(const int iterations, const element_type remaining_error)
      : iterations(iterations), remaining_error(remaining_error)
    {}
  };

  struct unknown_inversion_method_error
  {
    const char method;

    unknown_inversion_method_error(const char method)
      : method(method)
    {}
  };

  struct partition
  {
    matrix ul,
           ur,
           dl,
           dr;
    const bool tr_ul,
               tr_ur,
               & tr_dl,
               & tr_dr;

    // incomplete => symmetry -> transpose is ignored
    partition(const matrix & M, const bool transpose = false, const bool incomplete = false)
      : ul(M, partition_size(M.get_height()).first,  partition_size(M.get_width()).first,
              0,                                     0),
        ur(M, partition_size(M.get_height()).first,  partition_size(M.get_width()).second,
              0,                                     partition_size(M.get_width()).first),
        dl(M, partition_size(M.get_height()).second, partition_size(M.get_width()).first,
              partition_size(M.get_height()).first,  0),
        dr(M, partition_size(M.get_height()).second, partition_size(M.get_width()).second,
              partition_size(M.get_height()).first,  partition_size(M.get_width()).first),
        tr_ul(transpose && ! incomplete),
        tr_ur(transpose || incomplete),
        tr_dl(tr_ul),
        tr_dr(tr_ul)
    {
      if (incomplete)
        ur = dl;
      else if (transpose)
        dl.swap(ur);
    }
  };

  enum inversion_methods
  {
    LAPACK,
    LAPACK_SPD,
    NEWTON_SPD,
    STRASSEN_NEWTON,
    S_STRASSEN_NEWTON
  };

private:
  size_type row_length,
            offset;
  matrix_data_type data;

protected:
  void
  copy_to_new_array()
  {
    matrix new_matrix(get_height(), get_width());
    new_matrix.make_copy_of(*this);
    swap(new_matrix);
  }

  inline static void
  make_transpose(element_type * T, const size_type T_row_length,
      const size_type T_height, const size_type T_width,
      const element_type * M, const size_type M_row_length)
  {
    #if PARALLEL
      make_transpose_parallel(T, T_row_length, T_height, T_width, M, M_row_length, omp_get_num_procs());
    #else
      make_transpose_serial(T, T_row_length, T_height, T_width, M, M_row_length);
    #endif
  }

  static void
  make_transpose_parallel(element_type * T, const size_type T_row_length,
      const size_type T_height, const size_type T_width,
      const element_type * M, const size_type M_row_length,
      const int num_threads)
  {
    if (num_threads <= 1 || T_width < PARALLEL_TRANSPOSE_SIZE)
      make_transpose_serial(T, T_row_length, T_height, T_width, M, M_row_length);
    else
      #pragma omp parallel sections num_threads (std::min(4, num_threads))
      {
        // up-left
        #pragma omp section
        make_transpose_parallel(T, T_row_length,
            partition_size(T_height).first, partition_size(T_width).first,
            M, M_row_length,
            (num_threads + 3) / 4);
        // down-left
        #pragma omp section
        make_transpose_parallel(T + partition_size(T_height).first * T_row_length, T_row_length,
            partition_size(T_height).second, partition_size(T_width).first,
            M + partition_size(T_height).first, M_row_length,
            (num_threads + 3) / 4);
        // up-right
        #pragma omp section
        make_transpose_parallel(T + partition_size(T_width).first, T_row_length,
            partition_size(T_height).first, partition_size(T_width).second,
            M + partition_size(T_width).first * M_row_length, M_row_length,
            (num_threads + 3) / 4);
        // down-right
        #pragma omp section
        make_transpose_parallel(
            T + partition_size(T_height).first * T_row_length + partition_size(T_width).first, T_row_length,
            partition_size(T_height).second, partition_size(T_width).second,
            M + partition_size(T_height).first + partition_size(T_width).first * M_row_length, M_row_length,
            (num_threads + 3) / 4);
      }
  }

  static void
  make_transpose_serial(element_type * T, const size_type T_row_length,
      const size_type T_height, const size_type T_width,
      const element_type * M, const size_type M_row_length)
  {
    if (T_width < RECURSIVE_TRANSPOSE_SIZE)
    {
      for (size_type row = 0; row < T_height; ++row)
        for (size_type col = 0; col < T_width; ++col)
          T[row * T_row_length + col] = M[col * M_row_length + row];
    }
    else
    {
      // up-left
      make_transpose_serial(T, T_row_length,
          partition_size(T_height).first, partition_size(T_width).first,
          M, M_row_length);
      // down-left
      make_transpose_serial(T + partition_size(T_height).first * T_row_length, T_row_length,
          partition_size(T_height).second, partition_size(T_width).first,
          M + partition_size(T_height).first, M_row_length);
      // up-right
      make_transpose_serial(T + partition_size(T_width).first, T_row_length,
          partition_size(T_height).first, partition_size(T_width).second,
          M + partition_size(T_width).first * M_row_length, M_row_length);
      // down-right
      make_transpose_serial(
          T + partition_size(T_height).first * T_row_length + partition_size(T_width).first, T_row_length,
          partition_size(T_height).second, partition_size(T_width).second,
          M + partition_size(T_height).first + partition_size(T_width).first * M_row_length, M_row_length);
    }
  }

  inline static void
  make_symmetric(element_type * T, const size_type T_row_length, const size_type T_size)
  {
    #if PARALLEL
      make_symmetric_parallel(T, T_row_length, T_size, omp_get_num_procs());
    #else
      make_symmetric_serial(T, T_row_length, T_size);
    #endif
  }

  static void
  make_symmetric_parallel(element_type * T, const size_type T_row_length, const size_type T_size,
      const int num_threads)
  {
    if (num_threads <= 1 || T_size < PARALLEL_TRANSPOSE_SIZE)
      make_symmetric_serial(T, T_row_length, T_size);
    else
      #pragma omp parallel sections num_threads (3)
      {
        // up-left
        #pragma omp section
        make_symmetric_parallel(T, T_row_length, partition_size(T_size).first, (num_threads + 1) / 2);
        // up-right
        #pragma omp section
        make_transpose_parallel(T + partition_size(T_size).first, T_row_length,
            partition_size(T_size).first, partition_size(T_size).second,
            T + partition_size(T_size).first * T_row_length, T_row_length,
            num_threads);
        // down-right
        #pragma omp section
        make_symmetric_parallel(T + partition_size(T_size).first * T_row_length + partition_size(T_size).first,
            T_row_length, partition_size(T_size).second, (num_threads + 1) / 2);
      }
  }

  static void
  make_symmetric_serial(element_type * T, const size_type T_row_length, const size_type T_size)
  {
    if (T_size < RECURSIVE_TRANSPOSE_SIZE)
    {
      for (size_type row = 0; row < T_size - 1; ++row)
        for (size_type col = row + 1; col < T_size; ++col)
          T[row * T_row_length + col] = T[col * T_row_length + row];
    }
    else
    {
      // up-left
      make_symmetric_serial(T, T_row_length, partition_size(T_size).first);
      // up-right
      make_transpose_serial(T + partition_size(T_size).first, T_row_length,
          partition_size(T_size).first, partition_size(T_size).second,
          T + partition_size(T_size).first * T_row_length, T_row_length);
      // down-right
      make_symmetric_serial(T + partition_size(T_size).first * T_row_length + partition_size(T_size).first,
          T_row_length, partition_size(T_size).second);
    }
  }

  // *this = alpha * A * B + beta * *this
  void
  multiply_base(const matrix & A, const matrix & B,
  const ElementType alpha, const ElementType beta,
  const bool transpose_A, const bool transpose_B);

public:
  // ----- constructors and assignment -----

  explicit
  matrix(const size_type sidelength = 0)
    : matrix_b(sidelength),
      row_length(sidelength),
      offset(0),
      data(0)
  {
    if (sidelength != 0)
      data.reset(new element_type[sidelength * sidelength]);
  }

  explicit
  matrix(const size_type height, const size_type width)
    : matrix_b(height, width),
      row_length(width),
      offset(0),
      data()
  {
    if (height * width != 0)
      data.reset(new element_type[height * width]);
  }

  explicit
  matrix(const matrix & supermatrix,
      const size_type sidelength,
      const size_type from_row, const size_type from_col)
    : matrix_b(sidelength, supermatrix.is_symmetric() && from_row == from_col),
      row_length(supermatrix.get_row_length()),
      offset(supermatrix.offset + from_row * row_length + from_col),
      data(supermatrix.data)
  {
    assert(get_height() + from_row <= supermatrix.get_height());
    assert(get_width() + from_col <= supermatrix.get_width());
  }

  explicit
  matrix(const matrix & supermatrix,
      const size_type height, const size_type width,
      const size_type from_row, const size_type from_col)
    : matrix_b(height, width, supermatrix.is_symmetric() && from_row == from_col && height == width),
      row_length(supermatrix.row_length),
      offset(supermatrix.offset + from_row * row_length + from_col),
      data(supermatrix.data)
  {
    assert(get_height() + from_row <= supermatrix.get_height());
    assert(get_width() + from_col <= supermatrix.get_width());
  }

  ~matrix()
  {}

  // ----- informative -----

  using matrix_b::get_height;
  using matrix_b::get_width;
  using matrix_b::is_symmetric;
  using matrix_b::set_symmetric;

  inline void
  set_size(const size_type new_size)
  {
    if (new_size != get_height() || new_size != get_width())
      *this = matrix(new_size);
  }

  inline void
  set_size(const size_type new_height, const size_type new_width)
  {
    if (new_height != get_height() || new_width != get_width())
      *this = matrix(new_height, new_width);
  }

  inline const size_type &
  get_row_length() const
  {
    return row_length;
  }

  inline bool
  is_unique() const
  {
    return data.unique();
  }

  inline bool
  is_compact() const
  {
    return get_width() == get_row_length();
  }

  inline bool
  same_as(const matrix & M) const
  {
    return matrix_b::operator == (M) && offset == M.offset && data == M.data;
  }

  // ----- access -----

  inline const element_type *
  get_const_pointer(const size_type row = 0) const
  {
    return data.get() + offset + row * get_row_length();
  }

  inline element_type *
  get_pointer(const size_type row = 0)
  {
    return data.get() + offset + row * get_row_length();
  }

  inline const element_type &
  const_elem(const size_type row, const size_type col) const
  {
    return data[offset + row * row_length + col];
  }

  inline element_type &
  elem(const size_type row, const size_type col)
  {
    return data[offset + row * row_length + col];
  }

  // ----- manipulating -----

  matrix &
  swap(matrix & M)
  {
    matrix_b::swap(M);
    std::swap(row_length, M.row_length);
    std::swap(offset, M.offset);
    data.swap(M.data);
    return *this;
  }

  inline void
  detach()
  {
    if (! data.unique())
      copy_to_new_array();
  }

  inline void
  compact()
  {
    if (! is_compact())
      copy_to_new_array();
  }

  // asserts this and M of same height and width
  matrix &
  make_copy_of(const matrix & M)
  {
    if (! same_as(M))
    {
      assert(get_height() == M.get_height() && get_width() == M.get_width());
      set_symmetric(M.is_symmetric());

      #pragma omp parallel for if (get_height() >= PARALLEL_COPY_SIZE)
      for (size_type row = 0; row < get_height(); ++row)
        memcopy(get_pointer(row), M.get_const_pointer(row), get_width());
    }
    return *this;
  }

  // asserts this and M of swapped height and width
  // assumes this is not the same as M
  matrix &
  make_transpose_of(const matrix & M)
  {
    if (M.is_symmetric())
      make_copy_of(M);
    else
    {
      assert(get_height() == M.get_width() && get_width() == M.get_height());
      set_symmetric(false);
      make_transpose(get_pointer(), get_row_length(), get_height(), get_width(),
          M.get_const_pointer(), M.get_row_length());
    }
    return *this;
  }

  // makes this symmetric assuming the lower triangle is set with the right values
  // asserts this square
  matrix &
  make_symmetric()
  {
    assert(get_height() == get_width());
    make_symmetric(get_pointer(), get_row_length(), get_width());
    set_symmetric(true);
    return *this;
  }

  void
  load(std::istream & file)
  {
    size_type new_height, new_width;
    bool new_symmetric;
    file.read(reinterpret_cast<char*>(&new_height), sizeof(size_type));
    file.read(reinterpret_cast<char*>(&new_width), sizeof(size_type));
    file.read(reinterpret_cast<char*>(&new_symmetric), sizeof(bool));
    set_size(new_height, new_width);
    set_symmetric(new_symmetric);
    // if necessary read each line separately
    if (is_compact())
      file.read(reinterpret_cast<char*>(get_pointer()), get_height() * get_width() * sizeof (element_type));
    else
      for (size_type row = 0; row < get_height(); ++row)
        file.read(reinterpret_cast<char*>(get_pointer() + row * row_length), get_width() * sizeof (element_type));
  }

  void
  store(std::ostream & file) const
  {
    file.write(reinterpret_cast<const char*>(&(get_height())), sizeof(size_type));
    file.write(reinterpret_cast<const char*>(&(get_width())), sizeof(size_type));
    file.write(reinterpret_cast<const char*>(&(is_symmetric())), sizeof(bool));
    // if necessary read each line separately
    if (is_compact())
      file.write(reinterpret_cast<const char*>(get_const_pointer()), get_height() * get_width() * sizeof (element_type));
    else
      for (size_type row = 0; row < get_height(); ++row)
        file.write(reinterpret_cast<const char*>(get_const_pointer() + row * row_length), get_width() * sizeof (element_type));
  }

  // ----- operations -----

  matrix &
  operator += (const matrix & A)
  {
    return add(A);
  }

  // *this = alpha * *this + beta * A
  matrix &
  add(const matrix & A, const element_type alpha = 1, const element_type beta = 1)
  {
    assert(get_height() == A.get_height());
    assert(get_width() == A.get_width());
    #pragma omp parallel for if (get_height() >= PARALLEL_ADD_REDUCTION_SIZE)
    for (size_type row = 0; row < get_width(); ++row)
      for (size_type col = 0; col < get_width(); ++col)
        elem(row, col) = alpha * elem(row, col) + beta * A.const_elem(row, col);
    set_symmetric(is_symmetric() && A.is_symmetric());
    return *this;
  }

  inline matrix
  operator * (const matrix & right) const
  {
    matrix res(get_height(), right.get_width());
    return res.multiply(*this, right);
  }

  // *this = alpha * A * B + beta * *this
  matrix &
  multiply(const matrix & A, const matrix & B,
      const ElementType alpha = 1, const ElementType beta = 0,
      const bool transpose_A = false, const bool transpose_B = false);

  // lower_triangle(*this) = lower_triangle(alpha * A * B + beta * *this)
  // if (!incomplete_ok) upper_trangle(*this) = lower_triangle(*this)^T
  matrix &
  multiply_symmetric(const matrix & A, const matrix & B,
      const ElementType alpha = 1, const ElementType beta = 0,
      const bool transpose_A = false, const bool transpose_B = false,
      const bool incomplete_ok = false);

  matrix &
  multiply_stable(const matrix & A, const matrix & B,
      const ElementType alpha = 1, const ElementType beta = 0,
      const bool transpose_A = false, const bool transpose_B = false);

  // returns alpha * this * this^T
  matrix
  spd_square(const ElementType alpha = 1, const bool transpose = false) const;

  matrix
  invert(const bool is_symmetric_positive_definite = false,
      const std::string method = "L") const
  {
    return matrix_inverse<element_type>(*this, is_symmetric_positive_definite, method);
  }

  matrix &
  set_alpha_unity(const element_type alpha = 1)
  {
    #pragma omp parallel for if (get_height() >= PARALLEL_SET_SIZE)
    for (size_type row = 0; row < get_height(); ++row)
      for (size_type col = 0; col < get_width(); ++col)
        elem(row, col) = alpha * (row == col);
    set_symmetric(true);
    return *this;
  }

  matrix &
  subtract_from_alpha_unity(const element_type alpha)
  {
    #pragma omp parallel for if (get_height() >= PARALLEL_SET_SIZE)
    for (size_type row = 0; row < get_height(); ++row)
      for (size_type col = 0; col < get_width(); ++col)
        elem(row, col) = alpha * (row == col) - elem(row, col);
    set_symmetric(true);
    return *this;
  }

  // ----- reductions ------

  element_type
  one_norm() const
  {
    element_type max = 0;
    #pragma omp parallel if (get_height() >= PARALLEL_ADD_REDUCTION_SIZE)
    {
      element_type pr_max = 0;
      #pragma omp for nowait
      for (size_type row = 0; row < get_height(); ++row)
      {
        element_type row_sum = blas::asum(get_width(), get_const_pointer(row), 1);
        if (row_sum > pr_max)
          pr_max = row_sum;
      }
      if (pr_max > max)
        #pragma omp critical
        if (pr_max > max)
          max = pr_max;
    }
    return max;
  }

  // assumes get_height() <= get_width()
  element_type
  one_norm_diff_unity() const
  {
    element_type max = 0;
    #pragma omp parallel if (get_height() >= PARALLEL_ADD_REDUCTION_SIZE)
    {
      element_type pr_max = 0;
      #pragma omp for nowait
      for (size_type row = 0; row < get_height(); ++row)
      {
        element_type row_sum = 0;
        for (size_type col = 0; col < get_width(); ++col)
          row_sum += absolute(const_elem(row, col) - (row == col));
        if (row_sum > pr_max)
          pr_max = row_sum;
      }
      if (pr_max > max)
        #pragma omp critical
        if (pr_max > max)
          max = pr_max;
    }
    return max;
  }

  element_type
  inf_norm() const
  {
    boost::scoped_array<element_type> col_sums(new element_type[get_width()]);
    for (size_type col = 0; col < get_height(); ++col)
      col_sums[col] = 0;
    for (size_type row = 0; row < get_height(); ++row)
      for (size_type col = 0; col < get_height(); ++col)
        col_sums[col] += absolute(const_elem(row, col));
    element_type max = 0;
    for (size_type col = 0; col < get_height(); ++col)
      if (col_sums[col] > max)
        max = col_sums[col];
    return max;
  }

  element_type
  max_norm() const
  {
    element_type max = 0;
    #pragma omp parallel if (get_height() >= PARALLEL_MAX_REDUCTION_SIZE)
    {
      element_type pr_max = 0;
      #pragma omp for nowait
      for (size_type row = 0; row < get_height(); ++row)
        for (size_type col = 0; col < get_width(); ++col)
        {
          element_type a = absolute(const_elem(row, col));
          if (a > pr_max)
            pr_max = a;
        }
      if (pr_max > max)
        #pragma omp critical
        if (pr_max > max)
          max = pr_max;
    }
    return max;
  }

  element_type
  max_norm_diff(const matrix & M) const
  {
    element_type max = 0;
    #pragma omp parallel if (get_height() >= PARALLEL_DIFF_MAX_REDUCTION_SIZE)
    {
      element_type pr_max = 0;
      #pragma omp for nowait
      for (size_type row = 0; row < get_height(); ++row)
        for (size_type col = 0; col < get_width(); ++col)
        {
          element_type a = absolute(const_elem(row, col) - const_elem(row, col));
          if (a > pr_max)
            pr_max = a;
        }
      if (pr_max > max)
        #pragma omp critical
        if (pr_max > max)
          max = pr_max;
    }
    return max;
  }

  element_type
  max_norm_diff_unity() const
  {
    element_type max = 0;
    #pragma omp parallel if (get_height() >= PARALLEL_MAX_REDUCTION_SIZE)
    {
      element_type pr_max = 0;
      #pragma omp for nowait
      for (size_type row = 0; row < get_height(); ++row)
        for (size_type col = 0; col < get_width(); ++col)
        {
          element_type a = absolute(const_elem(row, col) - (row == col));
          if (a > pr_max)
            pr_max = a;
        }
      if (pr_max > max)
        #pragma omp critical
        if (pr_max > max)
          max = pr_max;
    }
    return max;
  }

  inline std::string
  to_string() const
  {
    std::ostringstream oss;
    oss << (*this);
    return oss.str();
  }
};

// if A is symmetric and B not to be transposed, only the lower triangle of A is read
// else if B is symmetric and A not to be transposed, only the lower triangle of B is read
template <typename ElementType>
matrix<ElementType> &
matrix<ElementType>::multiply(const matrix & A, const matrix & B,
    const ElementType alpha, const ElementType beta,
    bool transpose_A, bool transpose_B)
{
  #if EXPLOIT_SYMMETRIC_MULTIPLICATION && PARALLEL
    if (!omp_in_parallel() || get_height() < PARALLEL_MULTIPLY_SIZE)
  #endif
      multiply_base(A, B, alpha, beta, transpose_A, transpose_B);
  #if EXPLOIT_SYMMETRIC_MULTIPLICATION && PARALLEL
    else
    {
      // partition
      partition c(*this);
      partition a(A, transpose_A, A.is_symmetric()),
          #if AVOID_STRANGE_MULTIPLICATION_BUG
                b(B, transpose_B, B.is_symmetric() && ! A.is_symmetric());
          #else
                // somehow this produces wrong results:
                b(B, transpose_B, B.is_symmetric());
          #endif


      // multiply recursively
      #pragma omp parallel sections num_threads (4)
      {
        #pragma omp section
        {
          c.ul.multiply(a.ul, b.ul, alpha, beta, a.tr_ul, b.tr_ul);
          c.ul.multiply(a.ur, b.dl, alpha, 1, a.tr_ur, b.tr_dl);
        }
        #pragma omp section
        {
          c.dl.multiply(a.dr, b.dl, alpha, beta, a.tr_dr, b.tr_dl);
          c.dl.multiply(a.dl, b.ul, alpha, 1, a.tr_dl, b.tr_ul);
        }
        #pragma omp section
        {
          #if AVOID_STRANGE_MULTIPLICATION_BUG
            if (B.is_symmetric())
              c.dr.multiply(a.dl, b.dl, alpha, beta, a.tr_dl, ! b.tr_dl);
            else
          #endif
              c.dr.multiply(a.dl, b.ur, alpha, beta, a.tr_dl, b.tr_ur);
          c.dr.multiply(a.dr, b.dr, alpha, 1, a.tr_dr, b.tr_dr);
        }
        #pragma omp section
        {
          c.ur.multiply(a.ur, b.dr, alpha, beta, a.tr_ur, b.tr_dr);
          // the strange multiplication bug causes unpredictable results of the following line
          c.ur.multiply(a.ul, b.ur, alpha, 1, a.tr_ul, b.tr_ur);
        }
      }
    }
  #endif
  set_symmetric(false);
  return *this;
}

// if A is symmetric and B not to be transposed, only the lower triangle of A is read
// else if B is symmetric and A not to be transposed, only the lower triangle of B is read
template <typename ElementType>
matrix<ElementType> &
matrix<ElementType>::multiply_symmetric(const matrix & A, const matrix & B,
    const ElementType alpha, const ElementType beta,
    const bool transpose_A, const bool transpose_B,
    const bool incomplete_ok)
{
  #if EXPLOIT_SYMMETRIC_MULTIPLICATION
    if (get_height() < RECURSIVE_MULTIPLY_SIZE)
  #endif
      multiply_base(A, B, alpha, beta, transpose_A, transpose_B);
  #if EXPLOIT_SYMMETRIC_MULTIPLICATION
    else
    {
      // partition
      partition c(*this);
      partition a(A, transpose_A, A.is_symmetric()),
                b(B, transpose_B, B.is_symmetric());

      // multiply recursively
      #pragma omp parallel sections num_threads (4) if (get_height() >= PARALLEL_MULTIPLY_SIZE)
      {
        #pragma omp section
        {
          c.ul.multiply_symmetric(a.ul, b.ul, alpha, beta, a.tr_ul, b.tr_ul, true);
          c.ul.multiply_symmetric(a.ur, b.dl, alpha, 1, a.tr_ur, b.tr_dl, incomplete_ok);
        }
        #pragma omp section
        {
          c.dl.multiply(a.dr, b.dl, alpha, beta, a.tr_dr, b.tr_dl);
          c.dl.multiply(a.dl, b.ul, alpha, 1, a.tr_dl, b.tr_ul);
          if (!incomplete_ok)
            c.ur.make_transpose_of(c.dl);
        }
        #pragma omp section
        {
          c.dr.multiply_symmetric(a.dl, b.ur, alpha, beta, a.tr_dl, b.tr_ur, true);
          c.dr.multiply_symmetric(a.dr, b.dr, alpha, 1, a.tr_dr, b.tr_dr, incomplete_ok);
        }
      }
    }
  #endif
  set_symmetric(true);
  return *this;
}

// if A is symmetric and B not to be transposed, only the lower triangle of A is read
// else if B is symmetric and A not to be transposed, only the lower triangle of B is read
template <typename ElementType>
void
matrix<ElementType>::multiply_base(const matrix & A, const matrix & B,
    const ElementType alpha, const ElementType beta,
    bool transpose_A, bool transpose_B)
{
  assert(get_height() == ((transpose_A) ? A.get_width() : A.get_height()));
  assert(get_width() == ((transpose_B) ? B.get_height() : B.get_width()));
  assert(((transpose_A) ? A.get_height(): A.get_width()) == ((transpose_B) ? B.get_width() : B.get_height()));
  if (A.is_symmetric() && ! transpose_B)
    blas::symm(CblasRowMajor, CblasLeft, CblasLower,
        get_height(), get_width(),
        alpha, A.get_const_pointer(), A.get_row_length(), B.get_const_pointer(), B.get_row_length(),
        beta, get_pointer(), get_row_length());
  else if (B.is_symmetric() && ! transpose_A)
    blas::symm(CblasRowMajor, CblasRight, CblasLower,
        get_height(), get_width(),
        alpha, B.get_const_pointer(), B.get_row_length(), A.get_const_pointer(), A.get_row_length(),
        beta, get_pointer(), get_row_length());
  else
    blas::gemm(CblasRowMajor, blas::trans(transpose_A), blas::trans(transpose_B),
        get_height(), get_width(), (transpose_A) ? A.get_height(): A.get_width(),
        alpha, A.get_const_pointer(), A.get_row_length(), B.get_const_pointer(), B.get_row_length(),
        beta, get_pointer(), get_row_length());
}

template <typename ElementType>
matrix<ElementType> &
matrix<ElementType>::multiply_stable(const matrix & A, const matrix & B,
    const ElementType alpha, const ElementType beta,
    bool transpose_A, bool transpose_B)
{
  assert(! transpose_A && ! transpose_B);
  assert(get_height() == ((transpose_A) ? A.get_width() : A.get_height()));
  assert(get_width() == ((transpose_B) ? B.get_height() : B.get_width()));
  assert(((transpose_A) ? A.get_height(): A.get_width()) == ((transpose_B) ? B.get_width() : B.get_height()));

  #pragma omp parallel for
  for (size_type row = 0; row < get_height(); ++row)
  {
    boost::scoped_array<element_type> summands(new element_type [A.get_width() + 1]);

    for (size_type col = 0; col < get_width(); ++col)
    {
      for (size_type d = 0; d < A.get_width(); ++d)
        summands[d] = A.const_elem(row, d) * B.const_elem(d, col) * alpha;
      summands[A.get_width()] = const_elem(row, col) * beta;
      std::sort(summands.get(), summands.get() + A.get_width() + 1, absolute_greater<element_type>);
      element_type sum = 0;
      for (size_type d = 0; d < A.get_width() + 1; ++d)
        sum += summands[d];
      elem(row, col) = sum;
    }
  }
  return *this;
}

template <typename ElementType>
matrix<ElementType>
matrix<ElementType>::spd_square(const ElementType alpha, const bool transpose) const
{
  matrix S(transpose ? get_width() : get_height());

  // syrk and make_symmetric instead of general multiplication
  blas::syrk(CblasRowMajor, CblasLower, blas::trans(transpose),
      S.get_height(), transpose ? get_height() : get_width(),
      alpha, get_const_pointer(), get_row_length(), ElementType(0), S.get_pointer(), S.get_row_length());
  S.make_symmetric(); // */

  return S;
}

#include "matrix_inverse.hpp"

#endif /* MATRIX_HPP_ */










